package nintaco.files;

import java.io.*;
import nintaco.tv.*;
import static nintaco.util.BitUtil.*;
import static nintaco.mappers.NametableMirroring.*;
import static nintaco.tv.TVSystem.*;
import static nintaco.util.StreamUtil.*;

public class MutableNesFile extends NesFile implements Serializable, Cloneable {
  
  private static final long serialVersionUID = 0;
  
  public MutableNesFile(final DataInputStream in, final long fileSize) 
      throws Throwable {  
    this(in, fileSize, "", "");
  }
  
  public MutableNesFile(final DataInputStream in, final long fileSize, 
      final String entryFileName, final String archiveFileName) 
          throws Throwable {  
    super(in, fileSize, entryFileName, archiveFileName, false);
  }
  
  @Override
  protected void readROMs(final DataInputStream in, final long fileSize) 
      throws Throwable {
    
    trainer = new int[0];
    chrROM = new int[0];
    
    // Read the entire file body without decoding it.
    prgROM = new int[(int)(fileSize - header.length)];
    readBytes(in, prgROM);
  }
  
  public void write(String fileName) throws Throwable {
    write(new File(fileName));
  }
  
  public void write(File file) throws Throwable {
    try (DataOutputStream out = new DataOutputStream(new BufferedOutputStream(
        new FileOutputStream(file)))) {
      write(out);
    } 
  }
  
  public void write(DataOutputStream out) throws Throwable {

    writeString(out, HEADER_ID);

    out.write(prgRomPages);
    out.write(chrRomPages);
    
    int b6 = (mapperNumber & 0x0F) << 4;
    b6 = setBit(b6, 1, nonVolatilePrgRamPresent);
    b6 = setBit(b6, 2, trainer != null && trainer.length > 0);
    if (mirroring == VERTICAL) {
      b6 = setBit(b6, 0);
    } else if (mirroring == FOUR_SCREEN) {
      b6 = setBit(b6, 3);
    }
    out.write(b6);
    
    int b7 = mapperNumber & 0xF0;
    b7 = setBit(b7, 0, vsUniSystem);
    b7 = setBit(b7, 1, playchoice10);
    b7 = setBit(b7, 3, nes20Format);
    out.write(b7);

    if (nes20Format) {

      int b8 = (mapperNumber >> 8) & 0x0F;
      b8 |= (submapperNumber & 0x0F) << 4;
      out.write(b8);
      
      int b9 = (prgRomPages >> 8) & 0x0F;
      b9 |= (chrRomPages >> 4) & 0xF0;
      out.write(b9);
      
      int b10 = encodeRamSize(prgRamSize) & 0x0F;
      b10 |= (encodeRamSize(nonVolatilePrgRamSize) & 0x0F) << 4;
      out.write(b10);
      
      int b11 = encodeRamSize(chrRamSize) & 0x0F;
      b11 |= (encodeRamSize(nonVolatileChrRamSize) & 0x0F) << 4;
      out.write(b11); 
      
      int b12 = toBit(tvSystem == PAL);
      b12 |= setBit(b12, 1, anyTVSystem);
      out.write(b12);
      
      int b13 = vsPPU & 0x0F;
      b13 |= (vsMode & 0x0F) << 4;
      out.write(b13);

      out.write(new byte[2]);
    } else {
      
      out.write(prgRamSize / PRG_RAM_PAGE_SIZE);

      int b9 = 0;
      b9 = setBit(b9, 0, tvSystem == PAL);
      out.write(b9);
      
      out.write(new byte[6]);
    }

    if (isTrainerPresent()) {
      writeBytes(out, trainer);
    } 
    
    // Write the entire file body.
    writeBytes(out, prgROM);
  }
  
  private int encodeRamSize(int ramSize) {
    
    if (ramSize == 0) {
      return 0;
    }

    return log2(ramSize) - 6;
  } 
  
  public void setPrgROM(int[] prgROM) {
    this.prgROM = prgROM;
  }

  public void setChrROM(int[] chrROM) {
    this.chrROM = chrROM;
  }

  public void setTrainer(int[] trainer) {
    this.trainer = trainer;
  }

  public void setChrRomPages(int chrRomPages) {
    this.chrRomPages = chrRomPages;
  }

  public void setPrgRomPages(int prgRomPages) {
    this.prgRomPages = prgRomPages;
  }

  public void setPrgRamSize(int prgRamSize) {
    this.prgRamSize = prgRamSize;
  }

  public void setSubmapperNumber(int submapperNumber) {
    this.submapperNumber = submapperNumber;
  }

  public void setMirroring(int mirroring) {
    this.mirroring = mirroring;
  }

  public void setNonVolatilePrgRamSize(int NonVolatilePrgRamSize) {
    this.nonVolatilePrgRamSize = NonVolatilePrgRamSize;
  }

  public void setNonVolatileChrRamSize(int NonVolatileChrRamSize) {
    this.nonVolatileChrRamSize = NonVolatileChrRamSize;
  }

  public void setChrRamSize(int chrRamSize) {
    this.chrRamSize = chrRamSize;
  }

  public void setPrgRamPages(int prgRamPages) {
    this.prgRamPages = prgRamPages;
  }

  public void setTvSystem(TVSystem tvSystem) {
    this.tvSystem = tvSystem;
  }

  public void setVsPPU(int vsPPU) {
    this.vsPPU = vsPPU;
  }

  public void setVsMode(int vsMode) {
    this.vsMode = vsMode;
  }

  public void setNonVolatilePrgRamPresent(boolean sramPresent) {
    this.nonVolatilePrgRamPresent = sramPresent;
  }

  public void setPlaychoice10(boolean playchoice10) {
    this.playchoice10 = playchoice10;
  }

  public void setVsUnisystem(boolean vsUnisystem) {
    this.vsUniSystem = vsUnisystem;
  }

  public void setNes20Format(boolean nintaco20Format) {
    this.nes20Format = nintaco20Format;
  }

  public void setAnyTVSystem(boolean anyTVSystem) {
    this.anyTVSystem = anyTVSystem;
  }

  public void setMapperNumber(int mapperNumber) {
    this.mapperNumber = mapperNumber;
  }

  public MutableNesFile copy() {
    try {
      return (MutableNesFile)clone();
    } catch(Throwable t) {
      return null;
    }
  }
}
