package nintaco.files;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import nintaco.*;
import nintaco.cartdb.*;
import nintaco.tv.*;
import nintaco.mappers.*;
import nintaco.mappers.nintendo.vs.*;
import static nintaco.util.BitUtil.*;
import static nintaco.mappers.NametableMirroring.*;
import static nintaco.files.FileType.*;
import static nintaco.cartdb.CartDevices.*;
import static nintaco.tv.TVSystem.*;
import static nintaco.util.StreamUtil.*;
import static nintaco.util.StringUtil.*;

public class NesFile implements CartFile, Serializable, Cloneable {
  
  private static final long serialVersionUID = 0;
  
  public static final String HEADER_ID = "NES\u001A";
  
  protected static final String[] PAL_IDENTIFIERS 
      = { "(e)", "(europe)", "(pal)", "(f)", "(g)", "(i)" };
  
  protected static final String[] MAPPER_NAMES = {
    "NROM",                   //   0
    "MMC1",                   //   1
    "UNROM",                  //   2
    "CNROM",                  //   3
    "MMC3",                   //   4
    "MMC5",                   //   5
    "FFE Rev. A",             //   6
    "ANROM",                  //   7
    "",                       //   8
    "MMC2",                   //   9
    "MMC4",                   //  10
    "Color Dreams",           //  11
    "REX DBZ 5",              //  12
    "CPROM",                  //  13
    "REX SL-1632",            //  14
    "100-in-1",               //  15
    "BANDAI 24C02",           //  16
    "FFE Rev. B",             //  17
    "JALECO SS880006",        //  18
    "Namcot 106",             //  19
    "",                       //  20
    "Konami VRC2/VRC4 A",     //  21
    "Konami VRC2/VRC4 B",     //  22
    "Konami VRC2/VRC4 C",     //  23
    "Konami VRC6 Rev. A",     //  24
    "Konami VRC2/VRC4 D",     //  25
    "Konami VRC6 Rev. B",     //  26
    "CC-21 MI HUN CHE",       //  27
    "",                       //  28
    "",                       //  29
    "",                       //  30
    "",                       //  31
    "IREM G-101",             //  32
    "TC0190FMC/TC0350FMR",    //  33
    "IREM I-IM/BNROM",        //  34
    "Wario Land 2",           //  35
    "TXC Policeman",          //  36
    "PAL-ZZ SMB/TETRIS/NWC",  //  37
    "Bit Corp.",              //  38
    "",                       //  39
    "SMB2j FDS",              //  40
    "CALTRON 6-in-1",         //  41
    "BIO MIRACLE FDS",        //  42
    "FDS SMB2j LF36",         //  43
    "MMC3 BMC PIRATE A",      //  44
    "MMC3 BMC PIRATE B",      //  45
    "RUMBLESTATION 15-in-1",  //  46
    "NES-QJ SSVB/NWC",        //  47
    "TAITO TCxxx",            //  48
    "MMC3 BMC PIRATE C",      //  49
    "SMB2j FDS Rev. A",       //  50
    "11-in-1 BALL SERIES",    //  51
    "MMC3 BMC PIRATE D",      //  52
    "SUPERVISION 16-in-1",    //  53
    "",                       //  54
    "",                       //  55
    "SMB3 Pirate",            //  56
    "SIMBPLE BMC PIRATE A",   //  57
    "SIMBPLE BMC PIRATE B",   //  58
    "",                       //  59
    "SIMBPLE BMC PIRATE C",   //  60
    "20-in-1 KAISER Rev. A",  //  61
    "700-in-1",               //  62
    "Hello Kitty 255 in 1",   //  63
    "TENGEN RAMBO1",          //  64
    "IREM-H3001",             //  65
    "MHROM",                  //  66
    "SUNSOFT-FZII",           //  67
    "Sunsoft Mapper #4",      //  68
    "SUNSOFT-5/FME-7",        //  69
    "BANDAI KAMEN DISCRETE",  //  70
    "CAMERICA BF9093",        //  71
    "JALECO JF-17",           //  72
    "KONAMI VRC3",            //  73
    "TW MMC3+VRAM Rev. A",    //  74
    "KONAMI VRC1",            //  75
    "NAMCOT 108 Rev. A",      //  76
    "IREM LROG017",           //  77
    "Irem 74HC161/32",        //  78
    "AVE/C&E/TXC BOARD",      //  79
    "TAITO X1-005 Rev. A",    //  80
    "",                       //  81
    "TAITO X1-017",           //  82
    "YOKO VRC Rev. B",        //  83
    "",                       //  84
    "KONAMI VRC7",            //  85
    "JALECO JF-13",           //  86
    "74*139/74 DISCRETE",     //  87
    "NAMCO 3433",             //  88
    "SUNSOFT-3",              //  89
    "HUMMER/JY BOARD",        //  90
    "EARLY HUMMER/JY BOARD",  //  91
    "JALECO JF-19",           //  92
    "SUNSOFT-3R",             //  93
    "HVC-UN1ROM",             //  94
    "NAMCOT 108 Rev. B",      //  95
    "BANDAI OEKAKIDS",        //  96
    "IREM TAM-S1",            //  97
    "",                       //  98
    "VS Uni/Dual- system",    //  99
    "",                       // 100
    "",                       // 101
    "",                       // 102
    "FDS DOKIDOKI FULL",      // 103
    "",                       // 104
    "NES-EVENT NWC1990",      // 105
    "SMB3 PIRATE A",          // 106
    "MAGIC CORP A",           // 107
    "FDS UNROM BOARD",        // 108
    "",                       // 109
    "",                       // 110
    "",                       // 111
    "ASDER/NTDEC BOARD",      // 112
    "HACKER/SACHEN BOARD",    // 113
    "MMC3 SG PROT. A",        // 114
    "MMC3 PIRATE A",          // 115
    "MMC1/MMC3/VRC PIRATE",   // 116
    "FUTURE MEDIA BOARD",     // 117
    "TSKROM",                 // 118
    "NES-TQROM",              // 119
    "FDS TOBIDASE",           // 120
    "MMC3 PIRATE PROT. A",    // 121
    "",                       // 122
    "MMC3 PIRATE H2288",      // 123
    "",                       // 124
    "FDS LH32",               // 125
    "",                       // 126
    "Double Dragon pirate",   // 127
    "",                       // 128
    "",                       // 129
    "",                       // 130
    "",                       // 131
    "TXC/MGENIUS 22111",      // 132
    "SA72008",                // 133
    "MMC3 BMC PIRATE",        // 134
    "",                       // 135
    "TCU02",                  // 136
    "S8259D",                 // 137
    "S8259B",                 // 138
    "S8259C",                 // 139
    "JALECO JF-11/14",        // 140
    "S8259A",                 // 141
    "UNLKS7032",              // 142
    "TCA01",                  // 143
    "AGCI 50282",             // 144
    "SA72007",                // 145
    "SA0161M",                // 146
    "TCU01",                  // 147
    "SA0037",                 // 148
    "SA0036",                 // 149
    "S74LS374N",              // 150
    "",                       // 151
    "BANDAI 74161/7432",      // 152
    "BANDAI SRAM",            // 153
    "",                       // 154
    "",                       // 155
    "",                       // 156
    "BANDAI BARCODE",         // 157
    "",                       // 158
    "BANDAI 24C01",           // 159
    "SA009",                  // 160
    "",                       // 161
    "",                       // 162
    "",                       // 163
    "",                       // 164
    "",                       // 165
    "SUBOR Rev. A",           // 166
    "SUBOR Rev. B",           // 167
    "",                       // 168
    "",                       // 169
    "",                       // 170
    "",                       // 171
    "",                       // 172
    "",                       // 173
    "",                       // 174
    "",                       // 175
    "BMCFK23C",               // 176
    "",                       // 177
    "",                       // 178
    "",                       // 179
    "",                       // 180
    "",                       // 181
    "Super Donkey Kong",      // 182
    "",                       // 183
    "Atlantis no Nazo",       // 184
    "CNROM (CHR disabled)",   // 185
    "",                       // 186
    "",                       // 187
    "",                       // 188
    "Thunder Warrior",        // 189
    "",                       // 190
    "",                       // 191
    "TW MMC3+VRAM Rev. B",    // 192
    "NTDEC TC-112",           // 193
    "TW MMC3+VRAM Rev. C",    // 194
    "TW MMC3+VRAM Rev. D",    // 195
    "",                       // 196
    "",                       // 197
    "TW MMC3+VRAM Rev. E",    // 198
    "",                       // 199
    "",                       // 200
    "",                       // 201
    "",                       // 202
    "",                       // 203
    "",                       // 204
    "",                       // 205
    "NAMCOT 108 Rev. C",      // 206
    "TAITO X1-005 Rev. B",    // 207
    "",                       // 208
    "",                       // 209
    "",                       // 210
    "",                       // 211
    "",                       // 212
    "",                       // 213
    "",                       // 214
    "",                       // 215
    "",                       // 216
    "",                       // 217
    "",                       // 218
    "UNLA9746",               // 219
    "Debug Mapper",           // 220
    "UNLN625092",             // 221
    "",                       // 222
    "",                       // 223
    "",                       // 224
    "72-in-1",                // 225
    "BMC 22+20-in-1",         // 226
    "1200-in-1",              // 227
    "Action 52",              // 228
    "31-in-1",                // 229
    "BMC Contra+22-in-1",     // 230
    "20-in-1",                // 231
    "BMC QUATTRO",            // 232
    "BMC 22+20-in-1 RST",     // 233
    "BMC MAXI",               // 234
    "Golden Game 150-in-1",   // 235
    "",                       // 236
    "",                       // 237
    "UNL6035052",             // 238
    "",                       // 239
    "",                       // 240
    "",                       // 241
    "",                       // 242
    "S74LS374NA",             // 243
    "DECATHLON",              // 244
    "",                       // 245
    "FONG SHEN BANG",         // 246
    "",                       // 247
    "",                       // 248
    "",                       // 249
    "Time Diver Avenger",     // 250
    "",                       // 251
    "SAN GUO ZHI PIRATE",     // 252
    "DRAGON BALL PIRATE",     // 253
    "Ai Senshi Nicol",        // 254
    "",                       // 255    
  };
  
  public static final int PRG_ROM_PAGE_SIZE = 16384;
  public static final int PRG_RAM_PAGE_SIZE = 8192;
  public static final int CHR_ROM_PAGE_SIZE = 8192;
  
  protected transient int[] prgROM;
  protected transient int[] chrROM;
  protected transient int[] trainer;
  protected int[] header = new int[16];
  protected int chrRomPages;
  protected int prgRomPages;
  protected int prgRamPages;
  protected int prgRamSize;
  protected int mapperNumber;
  protected int originalMapperNumber;
  protected int submapperNumber;
  protected int originalSubmapperNumber;
  protected int mirroring;
  protected int originalMirroring;
  protected int nonVolatilePrgRamSize;
  protected int nonVolatileChrRamSize;
  protected int chrRamSize;
  protected TVSystem tvSystem;
  protected TVSystem originalTvSystem;  
  protected int vsPPU;
  protected int vsMode; 
  protected boolean trainerPresent;
  protected boolean chrRamPresent;
  protected boolean nonVolatilePrgRamPresent;
  protected boolean playchoice10;
  protected boolean vsUniSystem;
  protected boolean nes20Format;
  protected boolean anyTVSystem;
  protected String fileName;
  protected String entryFileName;
  protected String archiveFileName;
  protected String description;
  protected String mapperName;
  protected int prgRomCRC;
  protected int chrRomCRC;
  protected int fileCRC;
  protected int prgRomLength;
  protected int chrRomLength;
  protected Cart cart;
  protected VsGame vsGame;
  
  public NesFile(final DataInputStream in, final long fileSize, 
      final String entryFileName, final String archiveFileName) 
          throws Throwable {  
    this(in, fileSize, entryFileName, archiveFileName, true);
  }
  
  public NesFile(final DataInputStream in, final long fileSize, 
      final String entryFileName, final String archiveFileName, 
          final boolean modifyHeader) throws Throwable {
    
    this.entryFileName = entryFileName;
    this.archiveFileName = archiveFileName;
    this.fileName = FileUtil.getFileName(entryFileName)
        .toLowerCase(Locale.ENGLISH);
    
    readBytes(in, header);
    
    if (!compareStrings(HEADER_ID, header)) {
      throw new MessageException("Invalid or unknown file format.");
    }

    tvSystem = NTSC;
    prgRomPages = header[4];
    chrRomPages = header[5];

    mapperNumber = header[6] >> 4;
    if (getBitBool(header[6], 3)) {
      mirroring = FOUR_SCREEN;
    } else if (getBitBool(header[6], 0)) {
      mirroring = VERTICAL;
    } else {
      mirroring = HORIZONTAL;
    }
    originalMirroring = mirroring;
    trainerPresent = getBitBool(header[6], 2);
    nonVolatilePrgRamPresent = getBitBool(header[6], 1);
    nonVolatilePrgRamSize = nonVolatilePrgRamPresent ? 0x2000 : 0;

    nes20Format = (header[7] & 0x0C) == 0x08 && (header[7] & 0x03) != 0x03;
    
    if (!nes20Format) {
      if ((header[7] & 0x0C) != 0x00 || (header[7] & 0x03) == 0x03) {
        header[7] = header[8] = header[9] = 0x00;
      } else {
        for(int i = 10; i < 16; i++) {
          if (header[i] != 0x00) {
            header[7] = header[8] = header[9] = 0x00;
            break;
          }
        }
      }
    }
    
    mapperNumber |= header[7] & 0xF0;
    playchoice10 = getBitBool(header[7], 1);
    vsUniSystem = getBitBool(header[7], 0);

    if (nes20Format) {

      submapperNumber = header[8] >> 4;
      mapperNumber |= (header[8] & 0x0F) << 8;

      chrRomPages |= (header[9] & 0xF0) << 4;
      prgRomPages |= (header[9] & 0x0F) << 8;

      nonVolatilePrgRamSize = decodeRamSize(header[10] >> 4);
      nonVolatilePrgRamPresent = nonVolatilePrgRamSize > 0;
      prgRamSize = decodeRamSize(header[10] & 0x0F);
      prgRamPages = prgRamSize / PRG_RAM_PAGE_SIZE;

      nonVolatileChrRamSize = decodeRamSize(header[11] >> 4);
      chrRamSize = decodeRamSize(header[11] & 0x0F);      

      originalTvSystem = tvSystem = getBitBool(header[12], 0) ? PAL : NTSC;
      anyTVSystem = getBitBool(header[12], 1);

      vsPPU = header[13] & 0x0F;
      vsMode = header[13] >> 4;

    } else {

      prgRamPages = header[8];
      prgRamSize = PRG_RAM_PAGE_SIZE * prgRamPages;
            
      if (getBitBool(header[9], 0)) {
        originalTvSystem = tvSystem = PAL;
      } else {
        originalTvSystem = tvSystem = NTSC;
        if (modifyHeader) {
          for(String palIdentifier : PAL_IDENTIFIERS) {
            if (this.fileName.contains(palIdentifier)) {
              tvSystem = PAL;
              break;
            }
          }
        }
      }
    }

    readROMs(in, fileSize);

    originalMapperNumber = mapperNumber;  
    originalSubmapperNumber = submapperNumber;
    cart = CartDB.getCart(fileCRC);
    if (modifyHeader) {            
      vsGame = VsGame.getVsGame(this);
      if (vsGame != null) {
        if (vsGame.isDualSystemGame()) {
          throw new MessageException("Vs. DualSystem games are not supported.");
        }
        vsUniSystem = true;
        vsMode = vsGame.getMode();
        vsPPU = vsGame.getPPU();        
        cart = new Cart(fileCRC, vsGame.getMapper(), 0, NTSC, 
            vsGame.isZapperGame() ? Zapper : -1, vsGame.getMirroring());
      }
      if (cart != null && CartDB.isEnabled()) {
        mapperNumber = cart.getMapper();
        submapperNumber = cart.getSubmapper();
        if (cart.getMirroring() >= 0) {
          mirroring = cart.getMirroring();
        }
        tvSystem = cart.getTVSystem();
      }
    }

    mapperName = mapperNumber < MAPPER_NAMES.length 
        ? MAPPER_NAMES[mapperNumber] : "";
    final StringBuilder sb = new StringBuilder();
    if (archiveFileName != null) {
      appendLine(sb, "File name: %s <%s>", 
          FileUtil.getFileName(archiveFileName), 
              FileUtil.getFileName(entryFileName));
      appendLine(sb, "Directory: %s", 
          FileUtil.getDirectoryPath(archiveFileName));
    } else {
      appendLine(sb, "File name: %s", FileUtil.getFileName(entryFileName));
      appendLine(sb, "Directory: %s", 
          FileUtil.getDirectoryPath(entryFileName));
    }
    appendLine(sb, "File format: %s", nes20Format ? "NES 2.0" : "iNES");
    appendLine(sb, "File CRC: %08X%s", fileCRC,
        cart == null ? " (unknown)" : ""); 
    if (originalMapperNumber != mapperNumber) {
      appendLine(sb, "Mapper #: %d (modified, was %d)", mapperNumber,
          originalMapperNumber);
    } else {
      appendLine(sb, "Mapper #: %d", mapperNumber);
    }
    if (nes20Format || submapperNumber != 0) {
      if (originalSubmapperNumber != submapperNumber) {
        appendLine(sb, "Submapper #: %d (modified, was %d)", submapperNumber, 
            originalSubmapperNumber);
      } else {
        appendLine(sb, "Submapper #: %d", submapperNumber);
      }
    }
    appendLine(sb, "Mapper name: %s", mapperName);
    appendLine(sb, "Mirroring: %s", NametableMirroring.toString(mirroring));   
    append(sb, "PRG ROM size: %d bytes (%d pages x %d bytes)", 
        prgRomLength, prgRomPages, PRG_ROM_PAGE_SIZE);
    if (prgRomLength != prgROM.length) {
      append(sb, ", %d bytes (adjusted)", prgROM.length);
    }
    appendLine(sb);
    appendLine(sb, "PRG ROM CRC: %08X", prgRomCRC);
    append(sb, "CHR ROM size: %d bytes (%d pages x %d bytes)", 
        chrRomLength, chrRomPages, CHR_ROM_PAGE_SIZE);
    if (chrRomLength != chrROM.length) {
      append(sb, ", %d bytes (adjusted)", chrROM.length);
    }
    appendLine(sb);
    appendLine(sb, "CHR ROM CRC: %08X", chrRomCRC);
    appendLine(sb, "CHR RAM: %s", toYesNo(chrRamPresent));
    if (nes20Format) {
      appendLine(sb, "PRG RAM size: %d bytes", prgRamSize);
    } else {
      appendLine(sb, "PRG RAM size: %d bytes (%d pages x %d bytes)", 
          prgRamSize, prgRamPages, PRG_RAM_PAGE_SIZE);
    }
    appendLine(sb, "Non-Volatile PRG RAM: %s", 
        toYesNo(nonVolatilePrgRamPresent));    
    if (nes20Format) {
      appendLine(sb, "Non-Volatile PRG RAM size: %d bytes", 
          nonVolatilePrgRamSize);
      appendLine(sb, "Non-Volatile CHR RAM size: %d bytes", 
          nonVolatileChrRamSize);
      appendLine(sb, "CHR RAM size: %d bytes", chrRamSize);      
    }
    appendLine(sb, "TV System: %s", tvSystem);
    if (nes20Format) {
      appendLine(sb, "Any TV System: %s", toYesNo(anyTVSystem));
    }
    appendLine(sb, "Vs. Unisystem: %s", toYesNo(vsUniSystem));
    if (vsUniSystem) {
      appendLine(sb, "Vs. PPU: %s", VsPPU.toString(vsPPU));
      appendLine(sb, "Vs. Mode: %s", VsMode.toString(vsPPU));
    }
    appendLine(sb, "PlayChoice-10: %s", toYesNo(playchoice10));
    appendLine(sb, "Trainer: %s", toYesNo(trainerPresent));
    description = sb.toString();
  }
  
  protected void readROMs(final DataInputStream in, final long fileSize) 
      throws Throwable {
    
    if (trainerPresent) {
      trainer = new int[512];
      readFully(in, trainer, 0, trainer.length, null);
    } else {
      trainer = new int[0];
    }
    
    if (prgRomPages == 0) {
      prgRomPages = 0x100;
    }
    prgRomLength = PRG_ROM_PAGE_SIZE * prgRomPages;
    prgROM = new int[ceilBase2(prgRomLength)];
    final CRC32 crc = new CRC32();
    prgRomCRC = readFully(in, prgROM, 0, prgRomLength, crc);
    
    // King Neptune's Adventure
    if (mapperNumber == 1 && prgRomLength == 0x8000 
        && prgRomCRC == 0x3C6C8E6C) {
      prgRomPages <<= 1;
      prgRomLength <<= 1;
      final int[] rom = new int[0x10000];
      System.arraycopy(prgROM, 0, rom, 0, 0x8000);
      prgROM = rom;
      readFully(in, prgROM, 0x8000, 0x8000, crc);
      prgRomCRC = (int)crc.getValue();
    }

    chrRomLength = CHR_ROM_PAGE_SIZE * chrRomPages;
    if (chrRomLength == 0) {
      chrRamPresent = true;
      chrROM = new int[0];
      chrRomCRC = 0;
    } else {
      chrRamPresent = false;
      chrROM = new int[ceilBase2(chrRomLength)];    
      chrRomCRC = readFully(in, chrROM, 0, chrRomLength, crc);
    }
    fileCRC = (int)crc.getValue();
  }
  
  private int readFully(final DataInputStream in, final int[] data, 
      final int offset, final int length, final CRC32 crc) throws IOException {
    
    final CRC32 dataCRC = new CRC32();
    int bytesRead = 0;
    for(int i = 0; i < length; i++, bytesRead++) {
      int d = in.read();
      if (d < 0) {
        break;
      }
      data[offset + i] = d;
      dataCRC.update(d);
      if (crc != null) {
        crc.update(d);
      }
    }
    if (bytesRead > 0) {
      for(int i = offset + bytesRead; i < data.length; i++) {
        data[i] = data[i % bytesRead];
      }
    }
    return (int)dataCRC.getValue();
  }
  
  private int decodeRamSize(int value) {
    if (value == 0 || value == 15) {
      return 0;
    }
    return 0x40 << value;
  }

  @Override
  public int[] getPrgROM() {
    return prgROM;
  }

  @Override
  public int[] getChrROM() {
    return chrROM;
  }
  
  @Override
  public boolean isTrainerPresent() {
    return trainerPresent;
  }
  
  @Override
  public int getTrainerSize() {
    return trainer.length;
  }
  
  @Override
  public int[] getTrainer() {
    return trainer;
  }

  @Override
  public int getMapperNumber() {
    return mapperNumber;
  }

  public int getOriginalMapperNumber() {
    return originalMapperNumber;
  }

  @Override
  public int getSubmapperNumber() {
    return submapperNumber;
  }

  public int getOriginalSubmapperNumber() {
    return originalSubmapperNumber;
  }

  @Override
  public int getMirroring() {
    return mirroring;
  }

  public int getOriginalMirroring() {
    return originalMirroring;
  }

  public int getNonVolatilePrgRamSize() {
    return nonVolatilePrgRamSize;
  }

  public int getNonVolatileChrRamSize() {
    return nonVolatileChrRamSize;
  }

  @Override
  public int getChrRamSize() {
    return chrRamSize;
  }

  @Override
  public TVSystem getTvSystem() {
    return tvSystem;
  }

  public TVSystem getOriginalTvSystem() {
    return originalTvSystem;
  }

  public int getVsPPU() {
    return vsPPU;
  }

  @Override
  public int getVsMode() {
    return vsMode;
  }

  @Override
  public Cart getCart() {
    return cart;
  }

  @Override
  public VsGame getVsGame() {
    return vsGame;
  }

  @Override
  public boolean isChrRamPresent() {
    return chrRamPresent;
  }

  @Override
  public boolean isNonVolatilePrgRamPresent() {
    return nonVolatilePrgRamPresent;
  }

  public boolean isPlaychoice10() {
    return playchoice10;
  }

  public boolean isVsUniSystem() {
    return vsUniSystem;
  }

  public boolean isNes20Format() {
    return nes20Format;
  }

  public boolean isAnyTVSystem() {
    return anyTVSystem;
  }

  @Override
  public String getFileName() {
    return fileName;
  }

  public String getEntryFileName() {
    return entryFileName;
  }

  public String getArchiveFileName() {
    return archiveFileName;
  }

  public int getChrRomPages() {
    return chrRomPages;
  }
  
  public int getChrRomSize() {
    return CHR_ROM_PAGE_SIZE * chrRomPages;
  }
  
  public int[] getHeader() {
    return header;
  }
  
  public int getHeaderSize() {
    return header.length;
  }

  public int getPrgRomPages() {
    return prgRomPages;
  }
  
  public int getPrgRomSize() {
    return PRG_ROM_PAGE_SIZE * prgRomPages;
  }

  public int getPrgRamPages() {
    return prgRamPages;
  }

  public int getPrgRamSize() {
    return prgRamSize;
  }

  public String getMapperName() {
    return mapperName;
  }

  @Override
  public int getFileCRC() {
    return fileCRC;
  }

  public int getPrgRomCRC() {
    return prgRomCRC;
  }

  public int getChrRomCRC() {
    return chrRomCRC;
  }

  @Override
  public int getPrgRomLength() {
    return prgRomLength;
  }

  @Override
  public int getChrRomLength() {
    return chrRomLength;
  }

  @Override
  public int getFileType() {
    return NES;
  }
  
  @Override
  public String toString() {
    return description;
  }
  
  @Override
  public int[] getFileContents() {
    final int[] data = new int[getHeaderSize() + getTrainerSize() 
        + getPrgRomSize() + getChrRomSize()];
    System.arraycopy(header, 0, data, 0, getHeaderSize());
    System.arraycopy(trainer, 0, data, getHeaderSize(), getTrainerSize());
    System.arraycopy(prgROM, 0, data, getHeaderSize() + getTrainerSize(), 
        getPrgRomSize());
    System.arraycopy(chrROM, 0, data, getHeaderSize() + getTrainerSize() 
        + getPrgRomSize(), getChrRomSize());
    return data;
  }
  
  private void readObject(ObjectInputStream in) throws IOException, 
      ClassNotFoundException {
    in.defaultReadObject();
    prgROM = readByteArray(in);
    chrROM = readByteArray(in);
    trainer = readByteArray(in);    
  }
  
  private void writeObject(ObjectOutputStream out) throws IOException {
    out.defaultWriteObject();
    writeByteArray(out, prgROM);
    writeByteArray(out, chrROM);
    writeByteArray(out, trainer);
  }  
}
